import logging
import re
from string import ascii_lowercase

from meleager.models import URN, Alignment, Book, Edition, Passage, Text, Work

from ..helpers import DataProvider, build_ap_url, build_file_path, to_date, to_id
from .description import get_or_create_description
from .name import get_or_create_name

logger = logging.getLogger(__name__)


def create_entities(
    entity_data,
    user_data,
    author_data,
    language_id_to_model,
    keyword_data,
    language_to_edition_mapping,
    language_id_to_name,
    language_mapping,
):
    edition_id_to_model = {}
    edition_name = None

    passage_id_to_model = {}
    editions = set()
    text_id_to_model = {}
    work_id_to_model = {}
    alignments_id_to_model = {}
    comment_id_to_model = {}

    for entity in entity_data:
        endpoint = f"entities/{entity['id_entity']}/"

        entity = DataProvider(
            url=build_ap_url(endpoint, "v1"),
            file=build_file_path(endpoint),
            output_file=endpoint,
            save_local_file=True,
        )

        title = entity["title"]
        matchObj = re.match(
            r"([a-zA-Z\s]*)\s([0-9]*).([0-9]*[a-z]?)", title, re.M | re.I
        )
        if matchObj is None:
            logger.debug("SKIP | Invalid title for entity '%s'", title)
            continue

        work_desc, book, fragment = matchObj.groups()
        if not work_desc or not book or not fragment:
            logger.debug(
                "SKIP | Invalid title for entity '%s'. Regex match but got an invalid output: %s, book %s, fragment %s",
                title,
                work_desc,
                book,
                fragment,
            )
            continue

        try:
            book = int(book)
        except ValueError as e:
            logger.debug(
                "SKIP | Invalid book format, got %s but cannot parse int: %s",
                fragment,
                repr(e),
            )
            continue

        groups = re.match(
            "^(?P<fragment_id>\d+)\.?(?P<subfragment_id>[a-zA-Z]+)?$", fragment
        )
        if groups:
            fragment_id = groups["fragment_id"]
            subfragment_id = groups["subfragment_id"]

            if not subfragment_id:
                subfragment_id = ""
            try:
                fragment_id = int(fragment_id)
            except ValueError as e:
                logger.debug(
                    "Invalid fragment format, got %s but cannot parse int: %s",
                    fragment,
                    repr(e),
                )
        else:
            logger.debug("SKIP | Unable to parse fragment: %s", fragment)
            continue

        if entity["id_user"]["id_user"] not in user_data:
            logger.debug("SKIP | Invalid user_id", entity["id_user"]["id_user"])
            continue

        defaults = {
            "creator": user_data[entity["id_user"]["id_user"]],
            "last_editor": user_data[entity["id_user"]["id_user"]],
        }

        # TODO: add the perseus URN
        work_id = to_id(work_desc)
        if work_id in work_id_to_model:
            work_obj = work_id_to_model[work_id]
            Work.objects.filter(pk=work_obj.pk).update(**defaults)
            action = "UPDATE"
        else:
            work_obj = Work.objects.create(**defaults)
            work_id_to_model[work_id] = work_obj
            action = "NEW"

        logger.debug("%s | Work %d (%s)", action, work_obj.pk, work_id)

        book_obj, created = Book.objects.get_or_create(
            work=work_obj,
            number=book,
            defaults={
                "creator": user_data[entity["id_user"]["id_user"]],
                "last_editor": user_data[entity["id_user"]["id_user"]],
            },
        )

        action = ["UDPATE", "NEW"][created]
        logger.debug("%s | Book: %d (PK %d)", action, book_obj.number, book_obj.pk)

        passage_obj, created = Passage.objects.update_or_create(
            book=book_obj,
            fragment=fragment_id,
            sub_fragment=subfragment_id,
            defaults={
                "creator": user_data[entity["id_user"]["id_user"]],
                "last_editor": user_data[entity["id_user"]["id_user"]],
                "created_at": to_date(entity["createdAt"]),
                "updated_at": to_date(entity["updatedAt"]),
            },
        )

        action = ["UDPATE", "NEW"][created]
        logger.debug(
            "%s | Passage: %d.%d%s (PK %d)",
            action,
            book_obj.number,
            passage_obj.fragment,
            passage_obj.sub_fragment,
            passage_obj.pk,
        )

        for author in entity["authors"]:
            if author["id_author"] not in author_data:
                logger(
                    "SKIP | Invalid author id: %s (not defined in author_data)",
                    author["id_author"],
                )
                continue
            passage_obj.authors.add(author_data[author["id_author"]])
            logger.debug(
                "ADD | %s to Passage %d",
                author_data[author["id_author"]],
                passage_obj.pk,
            )

        passage_id_to_model[entity["id_entity"]] = passage_obj

        texts = []
        for version in entity["versions"]:
            language = language_id_to_model[version["id_language"]]

            name_obj = get_or_create_name(
                name=work_desc,
                language=language,
                creator=user_data[version["id_user"]],
                last_editor=user_data[version["id_user"]],
                created_at=to_date(version["createdAt"]),
                updated_at=to_date(version["updatedAt"]),
            )
            work_obj.names.add(name_obj)

            if version["edition"] is None:
                _edition_name = language_to_edition_mapping[
                    language_id_to_name[version["id_language"]]
                ]
            else:
                _edition_name = version["edition"]

            if _edition_name is None:
                logger.debug(
                    "SKIP | Invalid edition for entity id: %s. Edition is null and language '%s' does not indicate an edition",
                    str(entity["id_entity"]),
                    language,
                )
                continue

            edition_obj, created = Edition.objects.update_or_create(
                work=work_id_to_model[to_id(work_desc)],
                edition_type=Edition.EditionType.MANUSCRIPT,
                defaults={
                    "creator": user_data[version["id_user"]],
                    "last_editor": user_data[version["id_user"]],
                    "created_at": to_date(version["createdAt"]),
                    "updated_at": to_date(version["updatedAt"]),
                },
            )

            description = get_or_create_description(
                description=edition_name,
                language=language,
                creator=user_data[version["id_user"]],
                last_editor=user_data[version["id_user"]],
                created_at=to_date(version["createdAt"]),
                updated_at=to_date(version["updatedAt"]),
            )
            edition_obj.descriptions.add(description)

            edition_id_to_model[to_id(edition_name)] = edition_obj

            # text_id = to_id(version["text_translated"])
            text_id = version["id_entity_version"]

            defaults = {
                "passage": passage_obj,
                "creator": user_data[entity["id_user"]["id_user"]],
                "last_editor": user_data[entity["id_user"]["id_user"]],
                "created_at": to_date(entity["createdAt"]),
                "updated_at": to_date(entity["updatedAt"]),
            }

            if text_id not in text_id_to_model:
                text_obj = Text.objects.create(
                    **defaults, text=version["text_translated"]
                )
                text_id_to_model[text_id] = text_obj
                action = "NEW"
            else:
                text_obj = text_id_to_model[text_id]
                Text.objects.filter(pk=text_obj.pk).update(
                    **defaults, text=version["text_translated"]
                )
                action = "UPDATE"

            logger.debug("%s | Text (%s) %d", action, text_obj.text[:20], text_obj.pk)

            logger.debug(
                "ADD | Text %d to Passage %d",
                text_obj.pk,
                passage_id_to_model[version["id_entity"]].pk,
            )
            passage_id_to_model[version["id_entity"]].texts.add(text_obj)

        for alignment in entity["alignements"]:
            try:
                text1 = [
                    t
                    for t in entity["versions"]
                    if t["id_entity_version"] == alignment["source"]
                ][0]
                # text1_id = to_id(text1["text_translated"])
                text1_id = text1["id_entity_version"]
            except IndexError as e:
                text1_id = None
                # print("No text found with index:", alignment['source'], ", skipping alignements id:",
                #       alignment['id_align'])
                # continue

            try:
                text2 = [
                    t
                    for t in entity["versions"]
                    if t["id_entity_version"] == alignment["target"]
                ][0]
                # text2_id = to_id(text2["text_translated"])
                text2_id = text2["id_entity_version"]
            except IndexError as e:
                text2_id = None
                # print("No text found with index:", alignment['target'], ", skipping alignements id:",
                #       alignment['id_align'])
                # continue

            defaults = {
                "status": Text.Status.PUBLISHED,
                "text": text1["text_translated"],
                "edition": edition_id_to_model[to_id(edition_name)],
                "language": language_id_to_model[alignment["source_lang"]],
                "creator": user_data[text1["id_user"]],
                "last_editor": user_data[text1["id_user"]],
                "passage": passage_id_to_model[text1["id_entity"]],
                "created_at": to_date(text1["createdAt"]),
                "updated_at": to_date(text1["updatedAt"]),
            }
            if text1_id and text1_id not in text_id_to_model:
                text_obj = Text.objects.create(**defaults)
                text_id_to_model[text1_id] = text_obj
            elif text1_id:
                text_obj = text_id_to_model[text1_id]
                Text.objects.filter(pk=text_obj.pk).update(**defaults)

            defaults = {
                "status": Text.Status.PUBLISHED,
                "text": text2["text_translated"],
                "edition": edition_id_to_model[to_id(edition_name)],
                "language": language_id_to_model[alignment["target_lang"]],
                "creator": user_data[text2["id_user"]],
                "last_editor": user_data[text2["id_user"]],
                "passage": passage_id_to_model[text2["id_entity"]],
                "created_at": to_date(text2["createdAt"]),
                "updated_at": to_date(text2["updatedAt"]),
            }
            if text2_id and text2_id not in text_id_to_model:
                text_obj = Text.objects.create(**defaults)
                text_id_to_model[text2_id] = text_obj
            elif text2_id:
                text_obj = text_id_to_model[text2_id]
                Text.objects.filter(pk=text_obj.pk).update(**defaults)

            defaults = {
                "alignment_data": alignment["json"],
                "text_1": text_id_to_model[text1_id] if text1_id else None,
                "text_2": text_id_to_model[text2_id] if text2_id else None,
                "creator": user_data[alignment["id_user"]],
                "last_editor": user_data[alignment["id_user"]],
                "created_at": alignment["createdAt"],
                "updated_at": alignment["updatedAt"],
            }

            if alignment["id_align"] not in alignments_id_to_model:
                alignment_obj = Alignment.objects.create(**defaults)
                alignments_id_to_model[alignment["id_align"]] = alignment_obj
            else:
                alignment_obj = alignments_id_to_model[alignment["id_align"]]
                Alignment.objects.filter(pk=alignment_obj.pk).update(**defaults)

        for keyword in entity["keywords"]:
            try:
                passage_id_to_model[entity["id_entity"]].keywords.add(
                    keyword_data[keyword["id_keyword"]]
                )
            except KeyError as e:
                print(
                    "Invalid keyword at id",
                    keyword["id_keyword"],
                    repr(e),
                    ": skipping keyword on passage.",
                )

        external_refs = []
        for external_ref in entity["externalRef"]:
            creator = user_data[external_ref["id_user"]]

            if external_ref["id_language"] is None:
                language = language_mapping["français moderne"]
            else:
                language = external_ref["id_language"]

            external_ref_obj, created = URN.objects.update_or_create(
                urn=external_ref["url"],
                defaults={"creator": creator, "last_editor": creator},
            )

            description = get_or_create_description(
                external_ref["title"],
                language,
                creator,
                creator,
                to_date(external_ref["createdAt"]),
                to_date(external_ref["updatedAt"]),
            )

            external_ref_obj.descriptions.add(description)
            external_refs.append(external_ref_obj)

        passage_obj.external_references.add(*external_refs)

    return passage_id_to_model
