import datetime
import hashlib
import json
import os
import time
from pathlib import Path

import pytz
import requests

from django.conf import settings

AP_API_URL = "http://anthologia.ecrituresnumeriques.ca/api"

OBJECTS = (
    ("entities", "v1"),
    ("authors", "v1"),
    ("keywords", "v1"),
    ("users", "v1"),
    ("scholies", "v1"),
    ("cities", "v1"),
    ("notes", "v1"),
    ("images", "v1"),
    ("keyword_categories", "v2"),
    ("languages", "v2"),
)


def build_ap_url(endpoint, version):
    endpoint = endpoint.rstrip("/")
    return f"{AP_API_URL.rstrip('/')}/{version}/{endpoint}/"


def build_file_path(endpoint):
    endpoint, _ = os.path.splitext(str(endpoint).rstrip("/"))

    filepath = Path("/import") / f"{endpoint}.json"

    if not filepath.exists() or not os.access(filepath, os.R_OK):
        return None

    return filepath


OBJECTS_DICT = {
    endpoint: {
        "url": build_ap_url(endpoint, version),
        "file": build_file_path(endpoint),
    }
    for endpoint, version in OBJECTS
}


class DataProvider:
    def __new__(cls, object_name=None, load_from_file=False, **kwargs):
        if load_from_file:
            source = OBJECTS_DICT[object_name]["file"]
            if not source:
                source = OBJECTS_DICT[object_name]["url"]
                load_from_file = False
        elif "file" in kwargs:
            source = build_file_path(kwargs["file"])
            if not source:
                source = kwargs.get("url")
                load_from_file = False
        else:
            source = OBJECTS_DICT[object_name]["url"]
            load_from_file = False

        if load_from_file:
            with open(source, "r") as fp:
                data = json.load(fp)
        else:
            data = requests.get(source).json()

        if kwargs.get("save_local_file", False):
            save_file = None

            if object_name:
                save_file = build_file_path(object_name)
            elif "output_file" in kwargs:
                save_file = build_file_path(kwargs["output_file"])

            if save_file:
                os.makedirs(save_file.parent, exist_ok=True)
                with open(save_file, "w") as fp:
                    json.dump(data, fp)

        return data


def to_id(s):
    return int(hashlib.sha1(str(s).encode("utf8")).hexdigest(), 16) % (2 ** 16 - 1)


def to_date(s):
    if isinstance(s, str):
        return datetime.datetime(
            *(time.strptime(s, "%Y-%m-%dT%H:%M:%S.%fZ")[0:6]), tzinfo=pytz.utc
        )
    else:
        return s
