import re
from urllib.parse import urlparse

from django.apps import apps
from django.conf import settings
from django.db import models
from django.http import HttpResponseBadRequest
from django.urls import reverse

from .mixins import (
    AlternativeURNResourceMixin,
    DescriptableResourceMixin,
    EditableResourceMixin,
    ValidableResourceMixin,
)
from .urn import URN

class PassageManager(models.Manager):
    def get_by_natural_key(self, fragment, sub_fragment, *book_natural_key):
        book = apps.get_model("meleager", "Book").objects.get_by_natural_key(*book_natural_key)
        return self.get(fragment=fragment, sub_fragment=sub_fragment, book=book)

class Passage(
    EditableResourceMixin,
    ValidableResourceMixin,
    DescriptableResourceMixin,
    AlternativeURNResourceMixin,
    models.Model,
):
    """ Passage model """

    objects = PassageManager()

    authors = models.ManyToManyField("meleager.Author")
    city = models.ForeignKey(
        "meleager.City",
        related_name="passages",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )

    book = models.ForeignKey(
        "meleager.Book",
        related_name="passages",
        help_text="Book the passage is in",
        on_delete=models.CASCADE,
        db_index=True,
    )

    fragment = models.IntegerField(
        help_text="The fragment number in book.", db_index=True
    )

    sub_fragment = models.CharField(
        help_text="The sub fragment number in book.",
        blank=True,
        null=False,
        default="",
        max_length=3,
        db_index=True,
    )

    images = models.ManyToManyField(
        "meleager.Medium",
        related_name="mlgr_passages",
        help_text="The images associated to this resource.",
    )

    comments = models.ManyToManyField(
        "meleager.Comment",
        related_name="mlgr_passages",
        help_text="The comments of this resource.",
    )

    external_references = models.ManyToManyField(URN)
    internal_references = models.ManyToManyField("self", symmetrical=True)

    manuscripts = models.ManyToManyField(
        "meleager.Manuscript",
        help_text="The image of the manuscript for this passage.",
        related_name="mlgr_passages",
    )

    keywords = models.ManyToManyField(
        "meleager.Keyword",
        related_name="mlgr_passages",
        help_text="The keywords that tag this resource.",
    )

    def natural_key(self):
        return (self.fragment, self.sub_fragment) + self.book.natural_key()
    natural_key.dependencies = ['meleager.Work', 'meleager.Book']

    def __str__(self):
        return f"Passage {self.book.number}.{self.fragment}{self.sub_fragment}"
